"""
$ regi0 download
"""
import os

import click
import gdown

from ..util.config import CONFIG, CONFIG_PATH
from ..util.logger import LOGGER


@click.command(
    short_help="Download necessary data to run the regi0 geographic and taxonomic "
               "command line utilities."
)
@click.argument("url", type=str)
@click.argument("dst", type=str, default=None, required=False)
@click.option(
    "--quiet",
    default=False,
    is_flag=True,
    help="Silence information logging.",
    show_default=True
)
def download(url, dst, quiet):
    if not os.path.exists(CONFIG_PATH):
        raise Exception("Config file not found. Run `regi0 setup` first.")

    output_path = gdown.download(url, dst, quiet=quiet)

    if not quiet:
        LOGGER.info(f"Extracting data in {os.path.basename(output_path)}")
    extracted_paths = gdown.extractall(output_path)

    if not quiet:
        LOGGER.info(f"Updating config at {CONFIG_PATH}.")
    for path in extracted_paths:
        if os.path.isfile(path):
            path = os.path.abspath(path)
            name = os.path.splitext(os.path.basename(path))[0]
            if CONFIG.has_option("paths", name):
                CONFIG.set("paths", name, path)
        else:
            if path.endswith("txt/"):
                CONFIG.set("paths", "checklists", path)
    with open(CONFIG_PATH, "w") as config_file:
        CONFIG.write(config_file)



"""
Helper functions for the geographic module.
"""
import re

import numpy as np
import pandas as pd
import rasterio
from scipy import stats


def create_id_grid(
    xmin: float,
    ymin: float,
    xmax: float,
    ymax: float,
    resolution: float,
    crs: str = "epsg:4326",
) -> rasterio.io.DatasetWriter:
    """
    Creates an in-memory raster with a grid where each pixel has a unique
    ID. The unique IDs start at 0 in the upper left corner and increments
    from left to right and top to bottom. The max value for unique ID
    will be (height * width) - 1.

    Parameters
    ----------
    xmin
        Upper-left corner x coordinate.
    ymin
        Lower-right corner y coordinate.
    xmax
        Lower-left corner x coordinate.
    ymax
        Upper-left corner y coordinate.
    resolution
        Pixel resolution.
    crs
        Coordinate Reference System. Must be in the form epsg:code.

    Returns
    -------
    rasterio.io.DatasetWriter
        In-memory raster with unique IDs.

    Notes
    -----
    Coordinates and resolution should match with the reference system
    passed in crs.

    """
    height = np.ceil((ymax - ymin) / resolution).astype(int)
    width = np.ceil((xmax - xmin) / resolution).astype(int)
    transform = rasterio.transform.from_origin(xmin, ymax, resolution, resolution)
    arr = np.arange(height * width, dtype=np.uint32).reshape(height, width)

    memfile = rasterio.MemoryFile()
    grid = memfile.open(
        driver="MEM",
        height=height,
        width=width,
        count=1,
        crs=crs,
        transform=transform,
        dtype=rasterio.uint32,
    )
    grid.write(arr, 1)

    return grid


def extract_year(string: str) -> int:
    """
    Extracts a four-digit valid year (1900-2099) from a string. If there
    are multiple four-digit valid years on the string, the first
    occurrence is returned.

    Parameters
    ----------
    string
        String to extract the year from.

    Returns
    -------
    int
        Four-digit integer representing the year.

    """
    expr = r"(?:19|20)\d{2}"
    matches = re.findall(expr, string)
    if matches:
        year = matches[0]
    else:
        raise Exception("The string does not have any valid year.")

    return int(year)


def get_nearest_year(
    dates: pd.Series,
    reference_years: list,
    direction: str = "backward",
    round_unmatched=True,
) -> pd.Series:
    """
    Get the nearest year for each row in a Series from a given list of
    years.

    Parameters
    ----------
    dates
        Series with datetime-like objects.
    reference_years
        List of years to round each row to.
    direction
        Whether to search for prior, subsequent, or closest matches. Can
        be "backward", "nearest" or "forward". For more information go to:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.merge_asof.html
    round_unmatched
        Whether to round unmatched rows to the nearest year using a
        different direction than the one specified.

    Returns
    -------
    pd.Series
        Series with the nearest years.

    """
    if not dates.name:
        dates.name = "__date"

    reference_years = sorted(reference_years)
    has_date = dates.notna()
    years = pd.to_datetime(dates[has_date]).dt.year
    years = years.sort_values()

    dummy_df = pd.DataFrame({years.name: reference_years, "__year": reference_years})
    result = pd.merge_asof(years, dummy_df, on=years.name, direction=direction)["__year"]

    if round_unmatched:
        if direction == "backward":
            result[result.isna()] = min(reference_years)
        elif direction == "forward":
            result[result.isna()] = max(reference_years)

    # merge_asof result has a new index that has to be changed for the original. Also,
    # merge_asof does not work with NaN values. All dates that are NaNs are discarded in
    # the previous steps and then are added as NaNs to the result here.
    result.index = years.index
    nans = pd.Series(np.nan, index=dates[~has_date].index)
    result = result.append(nans)
    result = result.sort_index()

    return result


def is_outlier(
    values: np.ndarray, method: str = "iqr", threshold: float = 2.0
) -> np.ndarray:
    """
    Classifies outliers in an array of values.

    Parameters
    ----------
    values
        1D NumPy array of values.
    method
        Method to classify outliers. Can be "std", "iqr" or "zscore".
    threshold
        For the "std" method is the value to multiply the standard
        deviation with. For the "zscore" method, it is the lower limit
        (negative) and the upper limit (positive) to compare Z Scores to.

    Returns
    -------
    np.ndarray
        1D boolean NumPy array indicating whether values are outliers.

    """
    if method == "iqr":
        iqr = stats.iqr(values, nan_policy="omit")
        q1 = np.nanpercentile(values, 25)
        q3 = np.nanpercentile(values, 75)
        lower_limit = q1 - (1.5 * iqr)
        upper_limit = q3 + (1.5 * iqr)

    elif method == "std":
        std = np.nanstd(values)
        mean = np.nanmean(values)
        lower_limit = mean - (threshold * std)
        upper_limit = mean + (threshold * std)

    elif method == "zscore":
        values = stats.zscore(values, nan_policy="omit")
        lower_limit = -threshold
        upper_limit = threshold

    else:
        raise ValueError("`method` must be one of 'std', 'iqr', 'zscore'.")

    return (values < lower_limit) | (values > upper_limit)



from regi0.taxonomic.local import (
    get_checklist_fields,
    get_checklist_fields_multiple,
    is_in_checklist,
    is_in_checklist_multiple
)
from regi0.taxonomic.parsing import get_canonical_name
from regi0.taxonomic.web import gnr, iucn, speciesplus



"""
$ regi0 geographic
"""
import click
from regi0.geographic.local import (
    get_layer_field_historical,
    intersects_layer_historical,
    find_outliers,
    find_spatial_duplicates
)
from regi0.io import read_geographic_table, write_table
from regi0.verification import verify
from rasterstats import point_query

from ..options import geographic as opts
from ..util.config import CONFIG
from ..util.logger import LOGGER


@click.command(
    short_help="Performs a complete geographic verification on a set of records."
)
@click.argument("src", type=click.Path(exists=True))
@click.argument("dst", type=click.Path(exists=False))
@opts.crs
@opts.lon_col
@opts.lat_col
@opts.date_col
@opts.admin0_col
@opts.admin1_col
@opts.admin2_col
@opts.species_col
@opts.admin0_path
@opts.admin1_path
@opts.admin2_path
@opts.urban_path
@opts.dem_path
@opts.admin0_match
@opts.admin1_match
@opts.admin2_match
@opts.default_year
@opts.gridres
@opts.mark
@opts.drop
@opts.quiet
def geographic(
    src,
    dst,
    crs,
    lon_col,
    lat_col,
    date_col,
    admin0_col,
    admin1_col,
    admin2_col,
    species_col,
    admin0_path,
    admin1_path,
    admin2_path,
    urban_path,
    dem_path,
    admin0_match,
    admin1_match,
    admin2_match,
    default_year,
    gridres,
    mark,
    drop,
    quiet
):
    if not quiet:
        LOGGER.info(f"Reading records from {src}.")
    records = read_geographic_table(src, lon_col, lat_col, crs=crs)

    # 1. Removal of records with incomplete or missing coordinates.
    if not quiet:
        LOGGER.info("Removing missing or incomplete coordinates.")
    records = records.dropna(how="any", subset=[lon_col, lat_col])

    # 2. Administrative boundaries verification.
    if not quiet:
        LOGGER.info("Validating administrative boundaries.")
    admin_paths = [admin0_path, admin1_path, admin2_path]
    admin_fields = [admin0_match, admin1_match, admin2_match]
    admin_cols = [admin0_col, admin1_col, admin2_col]
    keys = ["admin0", "admin1", "admin2"]
    for path, field, col, key in zip(admin_paths, admin_fields, admin_cols, keys):
        expected, source = get_layer_field_historical(
            records,
            path,
            date_col,
            field,
            direction="backward",
            round_unmatched=True,
            default_year=default_year,
            return_source=True
        )
        flag_name = CONFIG.get("flagnames", key)
        suggested_name = CONFIG.get("suggestednames", key)
        records = verify(
            records, col, expected, flag_name, True, suggested_name, drop=drop
        )
        records[CONFIG.get("sourcenames", key)] = source

    # 3. Urban limits verification.
    if not quiet:
        LOGGER.info("Identifying records in urban areas.")
    intersects, source = intersects_layer_historical(
        records,
        urban_path,
        date_col,
        direction="nearest",
        default_year=default_year,
        return_source=True,
    )
    records[CONFIG.get("flagnames", "urban")] = intersects
    records[CONFIG.get("sourcenames", "urban")] = source
    if drop:
        records = records[~records[CONFIG.get("flagnames", "urban")]]

    # 4. Elevation consistency
    if not quiet:
        LOGGER.info("Extracting elevation values and identifying outliers.")
    values = point_query(records.geometry, dem_path, interpolate="nearest")
    records[CONFIG.get("valuenames", "elevation")] = values
    is_outlier = find_outliers(
        records, species_col, CONFIG.get("valuenames", "elevation"), method="iqr"
    )
    records[CONFIG.get("flagnames", "elevation")] = is_outlier
    if drop:
        records = records[~records[CONFIG.get("flagnames", "elevation")]]

    # 5. Spatial duplicates
    if not quiet:
        LOGGER.info("Identifying spatial duplicates.")
    is_duplicate = find_spatial_duplicates(records, species_col, gridres, mark=mark)
    records[CONFIG.get("flagnames", "spatialduplicate")] = is_duplicate
    if drop:
        records = records[~records[CONFIG.get("flagnames", "spatialduplicate")]]

    if not quiet:
        LOGGER.info(f"Saving results to {dst}.")
    records = records.drop(columns="geometry")
    write_table(records, dst, index=False)



"""
Creates a ConfigParser object with parameters to run $ regi0 geographic
"""

import configparser
import os
import pathlib

import appdirs

CONFIG = configparser.ConfigParser()

CONFIG_PATH = os.path.join(appdirs.user_config_dir("regi0"), "settings.ini")
if os.path.exists(CONFIG_PATH):
    CONFIG.read(CONFIG_PATH)
else:
    root = pathlib.Path(__file__).parent.parent
    template_config_file = root.joinpath("config").joinpath("settings.ini")
    CONFIG.read(template_config_file)



"""
I/O common functions.
"""
import pathlib

import geopandas as gpd
import pandas as pd


def read_geographic_table(
    fn: str,
    lon_col: str,
    lat_col: str,
    crs: str = "epsg:4326",
) -> gpd.GeoDataFrame:
    """
    Reads tabular data (csv, txt, xls or xlsx) and converts it to a
    GeoDataFrame.

    Parameters
    ----------
    fn
        Filename with extension. Can be a relative or absolute path.
    lon_col
        Name of the longitude column.
    lat_col
        Name of the latitude column.
    crs
        Coordinate reference system with the corresponding EPSG code.
        Must be in the form epsg:code.

    Returns
    -------
    gpd.GeoDataFrame
        GeoDataFrame with the records.

    """
    dtypes = {lon_col: float, lat_col: float}
    df = read_table(fn, dtype=dtypes)
    geometry = gpd.points_from_xy(df[lon_col], df[lat_col])
    gdf = gpd.GeoDataFrame(df, geometry=geometry, crs=crs)

    return gdf


def read_table(fn: str, **kwargs) -> pd.DataFrame:
    """
    Reads tabular data (csv, txt, xls or xlsx).

    Parameters
    ----------
    fn
        Filename with extension. Can be a relative or absolute path.
    **kwargs
        pandas read_csv, read_table and read_excel keyword arguments.

    Returns
    -------
    pd.DataFrame
        DataFrame with the tabular data.

    """
    ext = pathlib.Path(fn).suffix
    if ext == ".csv":
        df = pd.read_csv(fn, **kwargs)
    elif ext == ".txt":
        df = pd.read_table(fn, **kwargs)
    elif ext in (".xls", ".xlsx"):
        df = pd.read_excel(fn, **kwargs)
    else:
        raise ValueError("Input file extension is not supported.")

    return df


def write_table(df: pd.DataFrame, fn: str, **kwargs) -> None:
    """
    Writes tabular data (csv, txt, xls or xlsx) to disk.

    Parameters
    ----------
    df
        DataFrame to write to disk.
    fn
        Filename with extension. Can be a relative or absolute path.
    **kwargs
        Keyword arguments for pandas read_csv, read_table and read_excel
        functions.

    Returns
    -------
    None

    """
    ext = pathlib.Path(fn).suffix
    if ext == ".csv":
        df.to_csv(fn, **kwargs)
    elif ext in (".xls", ".xlsx"):
        df.to_excel(fn, **kwargs)
    else:
        raise ValueError("Input file extension is not supported.")



"""
General helper functions.
"""
import pandas as pd


def clean_text(s: pd.Series) -> pd.Series:
    """
    Removes special characters, leading, trailing and multiple whitespaces
    from text values in a pandas Series.

    Parameters
    ----------
    s
        Series to clean.

    Returns
    -------
    pd.Series
        Clean Series.

    """
    s = s.replace(r"\d+|[^\w\s]+", "", regex=True)
    s = s.replace(r"\s+", " ", regex=True)
    s = s.str.strip()

    return s


def standardize_text(s: pd.Series) -> pd.Series:
    """
    Standardizes text values by cleaning, converting to lowercase and
    removing accents.

    Parameters
    ----------
    s
        Series to standardize.

    Returns
    -------
    pd.Series
        Standardized Series.

    """
    s = clean_text(s)
    s = s.str.lower()
    s = s.str.normalize("NFKD").str.encode("ascii", errors="ignore").str.decode("utf-8")

    return s



"""
Options for $ regi0 taxonomic
"""
import click

from ..util.config import CONFIG

species_col = click.option(
    "--species-col",
    type=str,
    default=CONFIG.get("colnames", "species"),
    help="Species name column.",
    show_default=True
)

data_sources = click.option(
    "--data-sources",
    type=str,
    default=None,
    help="Data source IDs to pass to Global Names Resolver API.",
    show_default=True
)

checklists_path = click.option(
    "--checklists-path",
    type=click.Path(exists=True),
    default=CONFIG.get("paths", "checklists"),
    help="Path to a folder with the checklists.",
    show_default=True
)

fields = click.option(
    "--fields",
    type=str,
    default=None,
    help="Checklist field(s) to retrieve. Pass it multiple times for multiple fields.",
    multiple=True,
    show_default=True
)

drop = click.option(
    "--drop",
    default=CONFIG.getboolean("behaviour", "drop"),
    is_flag=True,
    help="Drop records with a positive flag.",
    show_default=True
)

quiet = click.option(
    "--quiet",
    default=CONFIG.getboolean("behaviour", "quiet"),
    is_flag=True,
    help="Silence information logging.",
    show_default=True
)



"""
Functions for local geographic verifications.
"""
import glob
import os
from typing import Union

import fiona
import geopandas as gpd
import pandas as pd
from rasterstats import point_query

from regi0.geographic import utils


def _historical(
    gdf: gpd.GeoDataFrame,
    others_path: str,
    date_col: str,
    direction: str = "nearest",
    round_unmatched: bool = False,
    default_year: str = "none",
    op: str = "intersection",
    field: str = None,
    return_source: bool = False
) -> Union[pd.Series, tuple]:
    """
    Checks whether records in gdf intersect with features in other or if
    column values from gdf match with other column values after a spatial
    join for historical data. Supposing one has two or more layers for
    different years in time, this function will match each record with
    of those layers depending on the date of collection. Then, for each
    layer, it will compare the respective records by executing one of the
    check_intersect or check_match functions.

    Parameters
    ----------
    gdf
        GeoDataFrame with records.
    others_path
        Folder with shapefiles or GeoPackage file with historical data.
        Shapefile names of GeoPackage layer names must have a four-digit
        year anywhere in order to extract it and match it with the records
        collection date.
    date_col
        Column name with the collection date.
    direction
        Whether to search for prior, subsequent, or closest years. Can be
         "backward", "nearest" or "forward".
    round_unmatched
        Whether to round unmatched rows to the nearest year using a
        different direction than the one specified.
    default_year
        Default year to take for records that do not have a collection
        date or whose collection data did not match with any year. Can be
        "last" for the most recent year in the historical data, "fist" for
        the oldest year in the historical data or "none" to skip a default
        year assignation. Keep in mind that records without a collection
        date won't be validated.
    op
        Operation to execute. Can be "intersection" to execute
        check_intersection or "match" to execute check_match.
    field
        Field to get from layers when `op` is "match".
    return_source
        Whether to return a column with layer source.

    Returns
    -------
    gpd.GeoDataFrame
        Original GeoDataFrame (gdf) with extra columns.

    """
    if os.path.isdir(others_path):
        layers = glob.glob(os.path.join(others_path, "*.shp"))
        if not layers:
            raise Exception("`others_path` must contain shapefiles.")
        input_type = "shp"
    else:
        if others_path.endswith(".gpkg"):
            layers = fiona.listlayers(others_path)
            input_type = "gpkg"
        else:
            raise ValueError("`others_path` must be a GeoPackage.")

    years = list(map(utils.extract_year, layers))
    historical_year = utils.get_nearest_year(
        gdf[date_col], years, direction=direction, round_unmatched=round_unmatched
    )
    if default_year == "last":
        historical_year = historical_year.fillna(max(years))
    elif default_year == "first":
        historical_year = historical_year.fillna(min(years))
    elif default_year == "none":
        pass
    else:
        raise ValueError("`default_year` must be either 'first', 'last' or 'none'.")

    result = pd.Series(index=gdf.index)
    if return_source:
        source = pd.Series(index=gdf.index)

    for year in historical_year.dropna().unique():
        layer = layers[years.index(year)]
        if input_type == "shp":
            other = gpd.read_file(layer)
            year_source = os.path.splitext(os.path.basename(layer))[0]
        elif input_type == "gpkg":
            other = gpd.read_file(others_path, layer=layer)
            year_source = layer

        mask = historical_year == year
        year_gdf = gdf[mask]
        if op == "intersection":
            year_result = intersects_layer(year_gdf, other)
        elif op == "match":
            year_result = get_layer_field(year_gdf, other, field)
        else:
            raise ValueError("`op` must be either 'intersection' or 'match'.")

        result.loc[mask] = year_result
        if return_source:
            source.loc[mask] = year_source

    if return_source:
        return result, source
    else:
        return result


def get_layer_field(
    gdf: gpd.GeoDataFrame, other: gpd.GeoDataFrame, field: str, op: str = "intersects",
) -> pd.Series:
    """
    Gets the corresponding values of a specific field by performing a
    spatial join between a GeoDataFrame with records and a GeoDataFrame
    representing a vector layer.

    Parameters
    ----------
    gdf
        GeoDataFrame with records.
    other
        GeoDataFrame with the target layer.
    field
        Name of the field to extract values from.
    op
        Spatial join operation accepted by the GeoDataFrame's sjoin
        method.

    Returns
    -------
    pd.Series
        Extracted values.

    """
    join = gpd.sjoin(gdf, other, how="left", op=op)

    return join[field]


def get_layer_field_historical(
    gdf: gpd.GeoDataFrame, others_path: str, date_col: str, field: str, **kwargs
) -> Union[pd.Series, tuple]:
    """
    Gets the corresponding values of a specific field by performing a
    spatial join between a GeoDataFrame with records and multiple
    historical vector layers.

    Parameters
    ----------
    gdf
        GeoDataFrame with records.
    others_path
        Path of a .gpkg file or a folder containing .shp files with the
        historical layers.
    date_col
        Name of the date column in `gdf` to match historical layers with.
    field
        Name of the field to extract values from. Must exist in all the
        historical layers.
    **kwargs
        Keyword arguments accepted by the _historical function.

    Returns
    -------
    values
        Extracted values.
    source
        Corresponding source.

    """
    return _historical(gdf, others_path, date_col, op="match", field=field, **kwargs)


def intersects_layer(gdf: gpd.GeoDataFrame, other: gpd.GeoDataFrame) -> pd.Series:
    """
    Checks whether records from `gdf` intersect any feature of `other`.

    Parameters
    ----------
    gdf
        GeoDataFrame with records.
    other
        GeoDataFrame with features to intersect records with.

    Returns
    -------
    pd.Series
        Boolean Series indicating whether each record intersects other.

    """
    other = other.copy()

    # Ideally, one could check if the elements of `gdf` intersect any of
    # the features of `other` with the following line:
    # gdf.intersects(other.geometry.unary_union)
    # While this works, depending on the complexity of the geometries of
    # `other` and the number of elements of `gdf`, the execution can be
    # considerably slow. A workaround is to create a new column with a
    # constant value (e.g. 1) in `other` and do a spatial join. Elements
    # of `gdf` that have a value for that column intersect any of the
    # features in `other`.
    other["__dummy"] = 1
    join = gpd.sjoin(gdf, other, how="left", op="intersects")
    intersects = join["__dummy"].notna()
    intersects.name = None

    intersects.loc[~gdf.is_valid] = pd.NA

    return intersects


def intersects_layer_historical(
    gdf: gpd.GeoDataFrame, others_path: str, date_col: str, **kwargs
) -> Union[pd.Series, tuple]:
    """
    Checks whether records from `gdf` intersect any feature of their
    corresponding historical vector layer.

    Parameters
    ----------
    gdf
        GeoDataFrame with records.
    others_path
        Path of a .gpkg file or a folder containing .shp files with the
        historical layers.
    date_col
        Name of the date column in `gdf` to match historical layers with.
    kwargs
        Keyword arguments accepted by the _historical function.

    Returns
    -------
    intersects
        Boolean Series indicating whether each record intersects other.
    source
        Corresponding source.
    """
    return _historical(gdf, others_path, date_col, op="intersection", **kwargs)


def find_outliers(
    gdf: gpd.GeoDataFrame,
    species_col: str,
    value_col: str,
    method: str = "std",
    threshold: float = 2.0
) -> pd.Series:
    """
    Finds outlier records based on values of a specific column.

    Parameters
    ----------
    gdf
        GeoDataframe with records.
    species_col
        Column name with the species name for each record.
    value_col
        Column name with values to find outliers from.
    method
        Method to find outliers. Can be "std" for Standard Deviation,
        "iqr" for Interquartile Range or "zscore" for Z Score.
    threshold
        For the "std" method is the value to multiply the standard
        deviation with. For the "zscore" method, it is the lower limit
        (negative) and the upper limit (positive) to compare Z Scores to.

    Returns
    -------
    pd.Series
        Boolean Series indicating whether values are outliers.

    """
    result = pd.Series(index=gdf.index, dtype=bool)
    for species in gdf[species_col].unique():
        mask = gdf[species_col] == species
        values = gdf.loc[mask, value_col]
        result.loc[mask] = utils.is_outlier(values, method, threshold)
        # Result for records that do not have a value is left empty. This
        # is done here instead of returning the respective nan values in
        # the is_outlier function because numpy cannot combine Boolean and
        # nan values in a single array (while pandas allows it).
        result.loc[mask & gdf[value_col].isna()] = pd.NA

    return result


def find_spatial_duplicates(
    gdf: gpd.GeoDataFrame,
    species_col: str,
    resolution: float,
    bounds: Union[list, tuple] = None,
    mark: str = "all"
) -> pd.Series:
    """
    Find records of the same species that are in the same cell of a
    specific grid.

    Parameters
    ----------
    gdf
        GeoDataFrame with records.
    species_col
        Column name with the species name for each record.
    resolution
        Grid resolution.
    bounds
        Grid bounds (xmin, ym, xmax, ymax). If no bounds are passed, the
        bounds from gdf will be taken.
    mark
        Which duplicates to mark. Can be "head", to mark all but the last,
        "tail" to mark all but the first, or "all".

    Returns
    -------
    pd.Series
        Boolean Series indicating whether records are spatial duplicates.

    Notes
    -----
    bounds and resolution should match gdf coordinate reference
    system.

    """
    gdf = gdf.copy()

    if not bounds:
        bounds = gdf.geometry.total_bounds
    grid = utils.create_id_grid(*bounds, resolution, gdf.crs.srs)
    ids = point_query(
        gdf, grid.read(1), affine=grid.transform, interpolate="nearest", nodata=-9999
    )
    gdf["__grid_id"] = ids

    subset = [species_col, "__grid_id"]
    if mark == "all":
        keep = False
    elif mark == "head":
        keep = "last"
    elif mark == "tail":
        keep = "first"
    else:
        raise ValueError("Mark must be one of 'head', 'tail' or 'all'.")
    result = gdf.duplicated(subset, keep=keep)

    # Result for records that do not have a grid ID is left empty.
    no_grid_id = gdf["__grid_id"].isna()
    result.loc[no_grid_id] = pd.NA

    return result



"""
Wrappers for GNR API calls and derived functions.

API documentation can be found at: http://resolver.globalnames.org/api
"""
from typing import Union

import numpy as np
import pandas as pd
import requests

from regi0.taxonomic.utils import expand_result

API_URL = "http://resolver.globalnames.org/name_resolvers.json"


def resolve(
    names: Union[list, pd.Series, str],
    data_source_ids: list = None,
    resolve_once: bool = False,
    best_match_only: bool = False,
    with_context: bool = False,
    with_vernaculars: bool = False,
    with_canonical_ranks: bool = False
) -> pd.DataFrame:
    """
    Receives a list of names and resolves each against the entire resolver
    database or against specific data sources using the Global Names
    Resolver (GNR) API. Underlying resolving and scoring algorithms are
    described at: http://resolver.globalnames.org/about

    Parameters
    ----------
    names
        List of species names to resolve.
    data_source_ids
        List of specific data sources IDs to resolve against. A list of
        all the available data sources and their IDs can be found at:
        http://resolver.globalnames.org/data_sources.
    resolve_once
        Find the first available match instead of matches across all data
        sources with all possible renderings of a name.
    best_match_only
        Returns just one result with the highest score.
    with_context
        Reduce the likelihood of matches to taxonomic homonyms. When True,
        a common taxonomic context is calculated for all supplied names
        from matches in data sources that have classification tree paths.
        Names out of determined context are penalized during score
        calculation.
    with_vernaculars
        Return 'vernacular' field to present common names provided by a
        data source for a particular match.
    with_canonical_ranks
        Returns 'canonical_form' with infraspecific ranks, if they are
        present.

    Returns
    -------
    pd.DataFrame
        DataFrame where rows are the result for each match.

    """
    if isinstance(names, str):
        names = [names]
    if data_source_ids is None:
        data_source_ids = []

    # Apparently, the GNR API does not accept Booleans so they need to be
    # converted to lowercase strings first.
    params = {
        "data": "\n".join(names),
        "data_source_ids": "|".join(data_source_ids),
        "resolve_once": str(resolve_once).lower(),
        "best_match_only": str(best_match_only).lower(),
        "with_context": str(with_context).lower(),
        "with_vernaculars": str(with_vernaculars).lower(),
        "with_canonical_ranks": str(with_canonical_ranks).lower()
    }

    try:
        response = requests.post(API_URL, json=params)
        response.raise_for_status()
    except requests.exceptions.HTTPError as err:
        raise Exception(f"Error calling Global Name Resolver API. {err}")

    data = response.json()["data"]

    # The pd.json_normalize() function does not work when record_path
    # is not found in every single item inside the list of elements
    # passed. In some cases, the GNR API returns items without this key,
    # so it needs to be added (including an empty dictionary) before
    # normalizing the result.
    for item in data:
        if "results" not in item:
            item["results"] = [{}]

    return pd.json_normalize(data, record_path="results", meta="supplied_name_string")


def get_classification(
    names: Union[list, pd.Series, str],
    add_supplied_names: bool = False,
    add_source: bool = False,
    expand: bool = True,
    **kwargs,
) -> pd.DataFrame:
    """
    Gets the complete classification of multiple scientific names using
    the Global Names Resolver.

    Parameters
    ----------
    names
        Scientific name(s) to get results for.
    add_supplied_names
        Add supplied scientific names column to the resulting DataFrame.
    add_source
        Add source column to the resulting DataFrame.
    expand
        Whether to expand result rows to match `names` size. If False, the
        number of rows will correspond to the number of unique names in
        `names`. Only has effect if best_match_only=True is passed.
    **kwargs
        Keyword arguments of the resolve function.

    Returns
    -------
    pd.DataFrame
        DataFrame with the ranks for each match.

    """
    if isinstance(names, (list, str)):
        names = pd.Series(names)

    unique_names = pd.Series(names.dropna().unique())
    result = resolve(unique_names, **kwargs)

    ranks = ["kingdom", "phylum", "class", "order", "family", "genus", "species"]
    df = pd.DataFrame(columns=ranks, index=result.index)
    rank_indices = result["classification_path_ranks"].str.split("|", expand=True)
    path_indices = result["classification_path"].str.split("|", expand=True)

    for rank in ranks:
        # The GNR API result might have duplicated ranks for one or more
        # items. Thus, duplicated ranks are removed and only the value for
        # first the appearance is kept.
        rank_idx = np.nonzero(rank_indices.values == rank)
        unique_idx = np.unique(rank_idx[0], return_index=True)[1]
        new_idx = np.column_stack(rank_idx)[unique_idx]
        rank_idx = tuple(new_idx.T)
        rank_paths = path_indices.values[rank_idx]
        df.loc[(rank_indices == rank).any(axis=1), rank] = rank_paths

    df = df.replace("", np.nan)
    if add_supplied_names:
        df["supplied_name"] = unique_names
    if add_source:
        df["source"] = result["data_source_title"]
    if kwargs.get("best_match_only"):
        if expand:
            df = expand_result(df, names)

    return df



"""
$ regi0 taxonomic
"""
import glob
import os

import click
import pandas as pd
from regi0.io import read_table, write_table
from regi0.taxonomic.local import get_checklist_fields_multiple
from regi0.taxonomic.web.gnr import get_classification
from regi0.verification import verify

from ..options import taxonomic as opts
from ..util.config import CONFIG
from ..util.logger import LOGGER


@click.command(
    short_help="Performs a complete taxonomic verification on a set of records."
)
@click.argument("src", type=click.Path(exists=True))
@click.argument("dst", type=click.Path(exists=False))
@opts.species_col
@opts.data_sources
@opts.checklists_path
@opts.fields
@opts.drop
@opts.quiet
def taxonomic(
    src,
    dst,
    species_col,
    data_sources,
    checklists_path,
    fields,
    drop,
    quiet
):
    if not quiet:
        LOGGER.info(f"Reading records from {src}.")
    records = read_table(src)

    if not quiet:
        LOGGER.info("Validating scientific names using GNR.")
    if data_sources:
        data_sources = data_sources.split(",")
    classification = get_classification(
        records[species_col],
        add_supplied_names=False,
        add_source=True,
        expand=True,
        best_match_only=True,
        data_source_ids=data_sources
    )
    species = classification["species"]
    flag_name = CONFIG.get("flagnames", "species")
    suggested_name = CONFIG.get("suggestednames", "species")
    records = verify(
        records,
        species_col,
        species,
        flag_name,
        add_suggested=True,
        suggested_name=suggested_name,
        drop=drop
    )
    records[CONFIG.get("sourcenames", "species")] = classification["source"]

    # For extracting new information based on the scientific names, it is
    # necessary to pass accepted scientific names. Hence, if the user a
    # new series is created with the combination of originally correct
    # names and the new suggested ones for those cases where the resolver
    # found a suggestion.
    accepted_names = records.loc[records[flag_name], species_col]
    suggested_names = records.loc[~records[flag_name], suggested_name]
    names = pd.concat([accepted_names, suggested_names]).sort_index()

    if checklists_path:
        filenames = glob.glob(os.path.join(checklists_path, "*"))
    if fields:
        fields = list(fields)
        if not quiet:
            LOGGER.info("Retrieving fields from local checklists.")
        result = get_checklist_fields_multiple(
            names,
            filenames,
            species_col,
            fields,
            add_supplied_names=False,
            expand=True,
            keep_first=True,
            add_source=True,
            source_name="source"
        )
        records = pd.concat([records, result], axis=1)

    if not quiet:
        LOGGER.info(f"Saving results to {dst}.")
    write_table(records, dst, index=False)



"""
$ regi0 setup
"""
import pathlib
import os
import shutil

import appdirs
import click


@click.command(
    short_help="Sets up the configuration files required to run the other CLI commands."
)
def setup():
    root = pathlib.Path(__file__).parent.parent
    template_config_folder = root.joinpath("config")
    target_conf_folder = appdirs.user_config_dir("regi0")
    if not os.path.exists(target_conf_folder):
        os.makedirs(target_conf_folder)

    for config_file in template_config_folder.glob("*.ini"):
        shutil.copy(config_file, target_conf_folder)



"""
Options for $ regi0 geographic
"""
import click

from ..util.config import CONFIG

crs = click.option(
    "--crs",
    type=str,
    default=CONFIG.get("misc", "crs"),
    help="Coordinate Reference System in the form epsg:code.",
    show_default=True
)

lon_col = click.option(
    "--lon-col",
    type=str,
    default=CONFIG.get("colnames", "longitude"),
    help="Longitude column.",
    show_default=True
)

lat_col = click.option(
    "--lat-col",
    type=str,
    default=CONFIG.get("colnames", "latitude"),
    help="Latitude column.",
    show_default=True
)

date_col = click.option(
    "--date-col",
    type=str,
    default=CONFIG.get("colnames", "date"),
    help="Collection date column.",
    show_default=True
)

admin0_col = click.option(
    "--admin0-col",
    type=str,
    default=CONFIG.get("colnames", "admin0"),
    help="Level 0 administrative division column (i.e. country).",
    show_default=True
)

admin1_col = click.option(
    "--admin1-col",
    type=str,
    default=CONFIG.get("colnames", "admin1"),
    help="Level 1 administrative division column (e.g. department or state).",
    show_default=True
)

admin2_col = click.option(
    "--admin2-col",
    type=str,
    default=CONFIG.get("colnames", "admin2"),
    help="Level 2 administrative division column (e.g. municipality or county).",
    show_default=True
)

species_col = click.option(
    "--species-col",
    type=str,
    default=CONFIG.get("colnames", "species"),
    help="Species name column.",
    show_default=True
)

admin0_path = click.option(
    "--admin0-path",
    type=click.Path(exists=True),
    default=CONFIG.get("paths", "admin0"),
    help="""
        Path to a .gpkg file or a folder with .shp files with the level 0
        administrative boundaries.
    """,
    show_default=True
)

admin1_path = click.option(
    "--admin1-path",
    type=click.Path(exists=True),
    default=CONFIG.get("paths", "admin1"),
    help="""
        Path to a .gpkg file or a folder with .shp files with the level 1
        administrative boundaries.
    """,
    show_default=True
)

admin2_path = click.option(
    "--admin2-path",
    type=click.Path(exists=True),
    default=CONFIG.get("paths", "admin2"),
    help="""
        Path to a .gpkg file or a folder with .shp files with the level 2
        administrative boundaries.
    """,
    show_default=True
)

urban_path = click.option(
    "--urban-path",
    type=click.Path(exists=True),
    default=CONFIG.get("paths", "urban"),
    help="""
        Path to a .gpkg file or a folder with .shp files with the urban/population
        centers polygons.
    """,
    show_default=True,
)

dem_path = click.option(
    "--dem-path",
    type=click.Path(exists=True),
    default=CONFIG.get("paths", "dem"),
    help="Path to a raster file with a Digital Elevation Model (DEM).",
    show_default=True,
)

admin0_match = click.option(
    "--admin0-match",
    type=str,
    default=CONFIG.get("matchfields", "admin0"),
    help="Field name to match records using admin0_col.",
    show_default=True
)

admin1_match = click.option(
    "--admin1-match",
    type=str,
    default=CONFIG.get("matchfields", "admin1"),
    help="Field name to match records using admin1_col.",
    show_default=True
)

admin2_match = click.option(
    "--admin2-match",
    type=str,
    default=CONFIG.get("matchfields", "admin2"),
    help="Field name to match records using admin2_col.",
    show_default=True
)

default_year = click.option(
    "--default-year",
    type=str,
    default=CONFIG.get("misc", "defaultyear"),
    help="""
        Default year to take for records that do not have a collection date or whose 
        collection data did not match with any year. Can be 'last' for the most recent 
        year in the historical data, 'first' for the oldest year in the historical data 
        or 'none' to ignore the verification on records without a date.
    """,
    show_default=True
)

gridres = click.option(
    "--gridres",
    type=float,
    default=CONFIG.getfloat("duplicates", "gridres"),
    help="""
        Resolution of the grid to identify spatial duplicated records. Units must be the 
        same as crs.
    """,
    show_default=True
)

mark = click.option(
    "--mark",
    type=str,
    default=CONFIG.get("duplicates", "mark"),
    help="What duplicates to mark. Can be 'head', 'tail' or 'all'.",
    show_default=True
)

drop = click.option(
    "--drop",
    default=CONFIG.getboolean("behaviour", "drop"),
    is_flag=True,
    help="Drop records with a positive flag.",
    show_default=True
)

quiet = click.option(
    "--quiet",
    default=CONFIG.getboolean("behaviour", "quiet"),
    is_flag=True,
    help="Silence information logging.",
    show_default=True
)



"""
Helper functions for the taxonomic module.
"""
import pandas as pd


def expand_result(df: pd.DataFrame, names: pd.Series) -> pd.DataFrame:
    """
    Expands `df` rows to match `names` size by duplicating rows for
    duplicated names.

    Parameters
    ----------
    df
        DataFrame to expand.
    names
        Series with species names.

    Returns
    -------
    Copy of `df` wih duplicated rows to match `names` size.
    """
    df = df.copy()
    names = names.copy()

    if not names.name:
        names.name = "__supplied_name"
    df["__name"] = names.dropna().unique()
    df = pd.merge(names, df, left_on=names.name, right_on="__name", how="left")
    df = df.drop(columns=[names.name, "__name"])

    return df



"""
Open Nomenclature qualifiers and their abbreviations. Taken from Sigovini
et al. (2016). See: https://doi.org/10.1111/2041-210X.12594
"""
QUALIFIERS = {
    "aberratio": [
        "ab."
    ],
    "adultus|i": [
        "ad."
    ],
    "affinis": [
        "aff."
    ],
    "animalia cetera": [
        "a.c."
    ],
    "cetera": [
        "c.",
        "cet."
    ],
    "conchylium|a": [
        "conch.",
        "shell"
    ],
    "confer": [
        "cf.",
        "cfr.",
        "conf.",
        "sp. cf."
    ],
    "ex affinis": [
        "ex aff."
    ],
    "ex grege": [
        "ex gr.",
        "gr."
    ],
    "exuvia|ae": [
        "ex."
    ],
    "familia genus species": [
        "fam. gen. sp."
    ],
    "familia nova": [
        "fam. n.",
        "fam. nov.",
        "n. fam.",
        "nov. fam."
    ],
    "female": [
        "f."
    ],
    "forma": [
        "f."
    ],
    "fragmentum|a": [
        "frag.",
        "fr."
    ],
    "genus novum": [
        "g. n.",
        "g. nov.",
        "gen. et sp. nov.",
        "gen. n.",
        "gen. nov.",
        "n. g.",
        "nov. gen et sp.",
        "nov. gen."
    ],
    "genus species": [
        "g. sp.",
        "gen. sp."
    ],
    "incertae sedis": [
        "inc. sed."
    ],
    "juvenile|s": [
        "juv"
    ],
    "larva|ae": [
        "l.",
        "lv."
    ],
    "male": [
        "m."
    ],
    "nova species": [
        "n. sp.",
        "nov. sp.",
        "nov. spec."
    ],
    "ovigerous females": [
        "ov.",
        "ovig."
    ],
    "pupa|ae": [
        "p."
    ],
    "secundum": [
        "sec"
    ],
    "sensu": [
        "sensu"
    ],
    "sensu amplo": [
        "s. a.",
        "s. ampl."
    ],
    "sensu lato": [
        "s.l",
        "s. lat."
    ],
    "sensu stricto": [
        "s.s.",
        "s. str."
    ],
    "species": [
        "sp.",
        "spp."
    ],
    "species affinis": [
        "sp. aff."
    ],
    "species aggregate": [
        "agg.",
        "sp. agg."
    ],
    "species complex": [
        "complex"
    ],
    "species dubia": [
        "sp. dub."
    ],
    "species group": [
        "gr.",
        "group",
        "sp. gr."
    ],
    "species incerta": [
        "?",
        "gen. inc.",
        "gen. et sp. inc.",
        "inc.",
        "sp. inc."
    ],
    "species indeterminabilis": [
        "ind.",
        "indet.",
        "sp. ind.",
        "sp. indet."
    ],
    "species indeterminata": [
        "ind.",
        "indet.",
        "sp. ind.",
        "sp. indet."
    ],
    "species inquirenda": [
        "sp. inq."
    ],
    "species near": [
        "nr.",
        "sp. nr."
    ],
    "species nova": [
        "sp. n.",
        "sp. nov.",
        "spec. nov."
    ],
    "species plurimae": [
        "sp. pl."
    ],
    "species proxima": [
        "prox.",
        "sp. prox."
    ],
    "stetit": [
        "stet."
    ],
    "subspecies": [
        "ssp.",
        "subsp."
    ],
    "subspecies nova": [
        "n. ssp.",
        "ssp. n.",
        "ssp. nov.",
        "subsp. n.",
        "subsp. nov."
    ],
    "valva|e": [
        "valva"
    ],
    "varietas": [
        "v.",
        "var."
    ]
}



"""
Functions for local taxonomic verifications.
"""
import pathlib
from typing import Union

import pandas as pd


from regi0.io import read_table


def get_checklist_fields(
    names: Union[list, pd.Series, str],
    checklist: pd.DataFrame,
    name_field: str,
    fields: Union[list, str],
    add_supplied_names: bool = False,
    expand: bool = True
) -> pd.DataFrame:
    """
    Retrieves values for one or multiple fields from a checklist given
    some species names.

    Parameters
    ----------
    names
        Series with species names.
    checklist
        DataFrame wih checklist information.
    name_field
        Name of the column in `checklist` with species names.
    fields
        List of fields (columns) to retrieve from `checklist`.
    add_supplied_names
        Whether to add `names` as an extra column in the result.
    expand
        Whether to expand result rows to match `names` size. If False,
        the number of rows will correspond to the number of unique names
        in `names`.

    Returns
    -------
    pd.DataFrame
        DataFrame with the values retrieved from `checklist`.

    """
    if isinstance(names, (list, str)):
        names = pd.Series(names)
    names.name = "supplied_name"
    if isinstance(fields, str):
        fields = [fields]

    result = pd.merge(
        names, checklist, how="left", left_on="supplied_name", right_on=name_field
    )
    present_fields = list(set(fields).intersection(checklist.columns))
    absent_fields = list(set(fields).difference(checklist.columns))
    result = result[present_fields + ["supplied_name"]]
    result[absent_fields] = pd.NA
    result = result[fields + ["supplied_name"]]

    if not expand:
        result = result.drop_duplicates("supplied_name", ignore_index=True)
    if not add_supplied_names:
        result = result.drop(columns="supplied_name")

    return result


def get_checklist_fields_multiple(
    names: Union[list, pd.Series, str],
    filenames: list,
    name_field: str,
    fields: Union[list, str],
    add_supplied_names: bool = False,
    expand: bool = True,
    keep_first: bool = True,
    add_source: bool = False,
    source_name: str = "source"
) -> pd.DataFrame:
    """
    Retrieves values for one or multiple fields from multiple checklists
    given some species names. If a species name is found on more than one
    checklist, only the field(s) values for one of them is kept.

    Parameters
    ----------
    names
        Series with species names.
    filenames
        List of checklist file names.
    name_field
        Name of the column in `checklist` with species names.
    fields
        List of fields (columns) to retrieve from `checklist`.
    add_supplied_names
        Whether to add `names` as an extra column in the result.
    expand
        Whether to expand result rows to match `names` size. If False,
        the number of rows will correspond to the number of unique names
        in `names`.
    keep_first
        Whether to keep the first match from a checklist or use the latest.
    add_source
        Whether to add the checklist name where the values were retrieved
        from.
    source_name
        Name of the column with the source.

    Returns
    -------
    pd.DataFrame
        DataFrame with the values retrieved from the checklists.

    """
    result = None
    for fn in filenames:
        checklist = read_table(fn)
        temp_result = get_checklist_fields(
            names, checklist, name_field, fields, add_supplied_names, expand
        )
        mask = temp_result[fields].notna().any(axis=1)
        if add_source:
            stem = pathlib.Path(fn).stem
            temp_result.loc[mask, source_name] = stem
        if result is None:
            result = temp_result
        else:
            if keep_first:
                mask = result[fields].isna().all(axis=1) & mask
            result[mask] = temp_result[mask]

    return result


def is_in_checklist(
    names: Union[list, pd.Series, str],
    checklist: pd.DataFrame,
    name_field: str,
    add_supplied_names: bool = False,
    expand: bool = True
) -> pd.DataFrame:
    """
    Checks whether some species names are found in a given checklist.

    Parameters
    ----------
    names
        Series with species names.
    checklist
        DataFrame wih checklist information.
    name_field
        Name of the column in `checklist` with species names.
    add_supplied_names
        Whether to add `names` as an extra column in the result.
    expand
        Whether to expand result rows to match `names` size. If False, the
        number of rows will correspond to the number of unique names in
        `names`.

    Returns
    -------
    pd.DataFrame
        DataFrame with a Boolean Series indicating whether `names` are
        present in `checklist`. If add_supplied_names=True is passed, the
        result will have an extra column.
    """
    if isinstance(names, (list, str)):
        names = pd.Series(names)
    names.name = "supplied_name"

    if not expand:
        names = names.drop_duplicates().dropna().reset_index(drop=True)
    result = names.isin(checklist[name_field])
    result.name = "in_checklist"

    result.loc[names.isna()] = pd.NA

    if add_supplied_names:
        result = pd.concat([result, names], axis=1)

    if isinstance(result, pd.Series):
        result = pd.DataFrame(result)

    return result


def is_in_checklist_multiple(
    names: Union[list, pd.Series, str],
    filenames: list,
    name_field: str,
    add_supplied_names: bool = False,
    expand: bool = True,
    keep_first: bool = True,
    add_source: bool = False,
    source_name: str = "source"
) -> Union[pd.DataFrame, pd.Series]:
    """
    Checks whether some species names are found in a multiple checklist.

    Parameters
    ----------
    names
        Series with species names.
    filenames
        List of checklist file names.
    name_field
        Name of the column in `checklist` with species names.
    add_supplied_names
        Whether to add `names` as an extra column in the result.
    expand
        Whether to expand result rows to match `names` size. If False,
        the number of rows will correspond to the number of unique names
        in `names`.
    keep_first
        Whether to keep the first match from a checklist or use the latest.
    add_source
        Whether to add the checklist name where the values were retrieved
        from.
    source_name
        Name of the column with the source.

    Returns
    -------
    pd.DataFrame
        DataFrame with a Boolean Series indicating whether `names` are
        present in the checklists. If add_supplied_names=True or
        add_source=True, the result will have extra columns.

    """
    result = None
    for fn in filenames:
        checklist = read_table(fn)
        temp_result = is_in_checklist(
            names, checklist, name_field, add_supplied_names, expand
        )
        mask = temp_result["in_checklist"].fillna(False)
        if add_source:
            stem = pathlib.Path(fn).stem
            temp_result.loc[mask, source_name] = stem
        if result is None:
            result = temp_result
        else:
            if keep_first:
                mask = ~result["in_checklist"].fillna(False) & mask
            result[mask] = temp_result[mask]

    return result



"""
CLI entry point.
"""
import click

from .commands.download import download
from .commands.geographic import geographic
from .commands.setup import setup
from .commands.taxonomic import taxonomic


@click.group()
def main():
    pass


main.add_command(download)
main.add_command(geographic)
main.add_command(setup)
main.add_command(taxonomic)



"""
Create logger object so CLI commands can access it.
"""

import logging
import logging.config
import os

import appdirs

LOGGER = None

config_path = os.path.join(appdirs.user_config_dir("regi0"), "logger.ini")
if os.path.exists(config_path):
    logging.config.fileConfig(config_path)
    LOGGER = logging.getLogger(__name__)



"""
Scientific name parsing functions.
"""
import pandas as pd

from regi0.taxonomic.constants.qualifiers import QUALIFIERS
from regi0.utils import clean_text


def get_canonical_name(names: pd.Series) -> pd.Series:
    """
    Extracts the canonical name (genus and specific epithet) of a Series
    of scientific names. It does this by removing special characters,
    numbers and Open Nomenclature qualifiers (such as aff. or cf.) and
    then taking the first two words.

    Parameters
    ----------
    names
        Series with the scientific names.

    Returns
    -------
    pd.Series
        Series with the extracted canonical names.

    """
    names = clean_text(names)
    names = names.str.capitalize()

    qualifiers = pd.Series(sum(QUALIFIERS.values(), []))
    qualifiers = qualifiers.str.replace(".", "", regex=False)
    qualifiers = pd.unique(qualifiers.str.split(" ", expand=True).stack())
    names = names.str.split(" ", expand=True).replace(qualifiers, pd.NA)
    names = names.apply(lambda x: x.str.cat(sep=" "), axis=1)
    names = clean_text(names)

    return names.str.split(" ").str[:2].str.join(" ")



"""
General verification functions.
"""
import pandas as pd
from rapidfuzz import fuzz

from regi0.utils import standardize_text


def match(
    left: pd.Series,
    right: pd.Series,
    preprocess: bool = False,
    fuzzy: bool = False,
    threshold: float = 0.8,
) -> pd.Series:
    """
    Compares values between two different Series to check if they match.

    Parameters
    ----------
    left
        Left Series.
    right
        Right Series.
    preprocess
        Whether to clean and standardize values before comparing them.
    fuzzy
        Whether to compare values using fuzzy logic.
    threshold
        Threshold to define equal values using fuzzy logic.

    Returns
    -------
    pd.Series
        Series with booleans indicating whether the values match.

    """
    if preprocess:
        left = standardize_text(left)
        right = standardize_text(right)

    if fuzzy:
        values = pd.DataFrame({"left": left, "right": right})
        score = values.apply(lambda row: fuzz.ratio(row["left"], row["right"]), axis=1)
        result = (score / 100) >= threshold
    else:
        result = left == right

    return result


def verify(
    df: pd.DataFrame,
    observed_col: str,
    expected: pd.Series,
    flag_name: str,
    add_suggested: bool = False,
    suggested_name: str = None,
    drop: bool = False,
    **kwargs
) -> pd.DataFrame:
    """
    Verifies that the values in a specific column from `df` match some
    expected values.

    Parameters
    ----------
    df
        DataFrame with values.
    observed_col
        Name of the column in `df` with the values to verify.
    expected
        Series with expected values. Has to match `df` length.
    flag_name
        Name of the resulting column indicating whether the observed
        values match the expected values.
    add_suggested
        Whether to add a column to the result with suggested values for
        those rows where the observed values do not match the expected
        values.
    suggested_name
        Name of the column for the suggested values. Only has effect when
        add_suggested=True is passed.
    drop
        Whether to drop the rows where the observed values do not match
        the expected values.
    kwargs
        Keyword arguments accepted by the match function.

    Returns
    -------
    pd.DataFrame
        Copy of `df` with extra columns.

    """
    df = df.copy()

    df[flag_name] = match(df[observed_col], expected, **kwargs)

    if add_suggested:
        df.loc[~df[flag_name], suggested_name] = expected.loc[~df[flag_name]]
    if drop:
        df = df[df[flag_name]]

    return df



"""
Wrappers for IUCN API calls.

API documentation can be found at: https://apiv3.iucnredlist.org/api/v3/docs
"""
from typing import Union
from urllib.parse import urljoin

import pandas as pd
import requests

from regi0.taxonomic.utils import expand_result

API_URL = "https://apiv3.iucnredlist.org/api/v3/"


def _request(url: str, token: str) -> requests.Response:
    """
    Creates a request for the IUCN API and handles HTTP exceptions.

    Parameters
    ----------
    url
        IUCN API endpoint.
    token
        IUCN API authentication token.

    Returns
    -------
    requests.Reponse
        Request response.

    """
    try:
        response = requests.get(url, params={"token": token})
        response.raise_for_status()
    except requests.exceptions.HTTPError as err:
        raise Exception(f"Error calling IUCN API. {err}")
    if "message" in response.json():
        msg = response.json()["message"]
        raise Exception(f"Error calling IUCN API. {msg}")

    return response


def get_country_occurrence(
    names: pd.Series,
    token: str,
    add_supplied_names: bool = False,
    add_source: bool = False,
    expand: bool = True
) -> pd.DataFrame:
    """
    Gets country occurrence and related information for multiple species
    using the IUCN API.

    Parameters
    ----------
    names
        Scientific name(s) to get results for.
    token
        IUCN API authentication token.
    add_supplied_names
        Add supplied scientific names column to the resulting DataFrame.
    add_source
        Add source column to the resulting DataFrame.
    expand
        Whether to expand result rows to match `names` size. If False,
        the number of rows will correspond to the number of unique names
        in `names`.

    Returns
    -------
    pd.DataFrame
        DataFrame with information about country occurrence for each name.

    """
    if isinstance(names, (list, str)):
        names = pd.Series(names)

    endpoint = urljoin(API_URL, "species/countries/name/")
    df = pd.DataFrame()

    unique_names = names.dropna().unique()
    for name in unique_names:
        response = _request(urljoin(endpoint, name), token)
        if response.json().get("result"):
            temp_df = pd.DataFrame(response.json()["result"])
            result = pd.Series({
                field: values for field, values in zip(temp_df.columns, temp_df.T.values)
            })
        else:
            result = pd.Series([], dtype="object")
        df = df.append(pd.Series(result), ignore_index=True)

    if add_supplied_names:
        df["supplied_name"] = unique_names
    if add_source:
        df["source"] = "IUCN"
    if expand:
        df = expand_result(df, names)

    return df


def get_species_info(
    names: Union[list, pd.Series, str],
    token: str,
    add_supplied_names: bool = False,
    add_source: bool = False,
    expand: bool = True
) -> pd.DataFrame:
    """
    Gets IUCN category and miscellaneous information for multiple species
    using the IUCN API.

    Parameters
    ----------
    names
        Scientific name(s) to get results for.
    token
        IUCN API authentication token.
    add_supplied_names
        Add supplied scientific names column to the resulting DataFrame.
    add_source
        Add source column to the resulting DataFrame.
    expand
        Whether to expand result rows to match `names` size. If False,
        the number of rows will correspond to the number of unique names
        in `names`.

    Returns
    -------
    pd.DataFrame
        DataFrame with IUCN species information for each name.

    """
    if isinstance(names, (list, str)):
        names = pd.Series(names)

    endpoint = urljoin(API_URL, "species/")
    df = pd.DataFrame()

    unique_names = names.dropna().unique()
    for name in unique_names:
        response = _request(urljoin(endpoint, name), token)
        if response.json().get("result"):
            result = pd.Series(response.json()["result"][0])
        else:
            result = pd.Series([], dtype="object")
        df = df.append(pd.Series(result), ignore_index=True)

    if add_supplied_names:
        df["supplied_name"] = unique_names
    if add_source:
        df["source"] = "IUCN"
    if expand:
        df = expand_result(df, names)

    return df



"""
Wrappers for Species+/CITES checklist API calls.

API documentation can be found at: http://api.speciesplus.net/documentation
"""
from typing import Union
from urllib.parse import urljoin

import pandas as pd
import requests

from regi0.taxonomic.utils import expand_result

API_URL = "https://api.speciesplus.net/api/v1/"


def _request(url: str, token: str, params: dict) -> requests.Response:
    """
    Creates a request for the Species+/CITES checklist API and handles
    HTTP exceptions.

    Parameters
    ----------
    url
        Species+/CITES checklist API endpoint.
    token
        Species+/CITES checklist API authentication token.
    params
        Request parameters.

    Returns
    -------
    requests.Response
        Request response.

    """
    headers = {"X-Authentication-Token": token}
    try:
        response = requests.get(url, params=params, headers=headers)
        response.raise_for_status()
    except requests.exceptions.HTTPError as err:
        raise Exception(f"Error calling Species+ API. {err}")

    return response


def get_taxon_concepts(
    names: Union[list, pd.Series, str],
    token: str,
    add_supplied_names: bool = False,
    add_source: bool = False,
    expand: bool = True
) -> pd.DataFrame:
    """
    Calls the taxon_concepts endpoint of the Species+/CITES checklist API
    for multiple scientific names.

    Parameters
    ----------
    names
        Scientific name(s) to get results for.
    token
        Species+/CITES checklist API authentication token.
    add_supplied_names
        Add supplied scientific names column to the resulting DataFrame.
    add_source
        Add source column to the resulting DataFrame.
    expand
        Whether to expand result rows to match `names` size. If False,
        the number of rows will correspond to the number of unique names
        in `names`.

    Returns
    -------
    pd.Series
        Series with the corresponding cites listings.

    """
    if isinstance(names, (list, str)):
        names = pd.Series(names)

    endpoint = urljoin(API_URL, "taxon_concepts")
    df = pd.DataFrame()

    unique_names = names.dropna().unique()
    for name in unique_names:
        response = _request(endpoint, token, {"name": name})
        if response.json().get("taxon_concepts"):
            result = pd.DataFrame(response.json()["taxon_concepts"])
        else:
            result = pd.Series([], dtype="object")
        df = df.append(result, ignore_index=True)

    if add_supplied_names:
        df["supplied_name"] = unique_names
    if add_source:
        df["source"] = "Species+/CITES"
    if expand:
        df = expand_result(df, names)

    return df
